import pickle

from .. import Transformer

from spintop.models import (
    SpintopTestRecordCollection
)

class TestFormatTransformer(Transformer):
    
    def __call__(self, data):
        # TODO if cache is desired, add it here.
        data = self.collect_fn(data)
        data = self.transform_fn(data)
        return data
        
    def collect_fn(self, data):
        return data
            
    def transform_fn(self, data):
        return data


from spintop.models.collection import SpintopTestRecordView

def test_default_primitive_test_record(flat_test_record):
    view = SpintopTestRecordView()
    
    result = view.apply(flat_test_record, flatten_dict=False)
    
    assert 'test_id' in result
    
def test_tags_default_primitive_test_record(flat_test_record):
    view = SpintopTestRecordView()
    
    flat_test_record.add_tag('tag1')
    
    result = view.apply(flat_test_record, flatten_dict=False)
    
    assert result['test_id']['tags']['tag1']


import random
import time
import datetime

from spintop.models import SpintopTreeTestRecordBuilder
from .base import Generator

def normal(mu, sigma):
    return random.normalvariate(mu, sigma)

def random_failure(failure_odd):
    return random.uniform(0, 1.0) < failure_odd

base_mu_gen = lambda index: 5+(index*0.25)
base_sigma_gen = lambda index: 2
test_name_gen = lambda index: 'basicgen_test'
dut_id_gen = lambda index: 'dut%d' % ((index%25) if 17 < (index%25) < 20 else index % 17)
duration_gen = lambda index: 1 if (index%25) == 10 or (index%25) == 17 else 0.2


class RandomTestGenerator(Generator):
    def generate(self, count=25):
        builder = SpintopTreeTestRecordBuilder()
        
        start_datetime = datetime.datetime.now()
        for index in range(count):
            duration = duration_gen(index)
            fails = random_failure(0.1)
            builder.set_top_level_information(
                start_datetime=start_datetime,
                dut_id=dut_id_gen(index),
                testbench_name=test_name_gen(index),
                duration=duration,
                outcome=dict(
                    is_pass=not fails
                )
            )
            start_datetime += datetime.timedelta(seconds=duration)
            yield builder.build()


import re
from .internal import FeatureRecord

class Query():
    model_type = FeatureRecord

    def __init__(self):
        # The field key must equal exactly value, or re.search if a compiled regex
        self._value_equals = dict()

        # The field of type list named key must contain value
        self._list_contains = dict()

        # The field named key must equal one of the sub value in the list value
        self._value_equals_one_of = dict()

    @property
    def value_equals(self):
        return self._value_equals

    @property
    def list_contains(self):
        return self._list_contains

    @property
    def value_equals_one_of(self):
        return self._value_equals_one_of

    def name_regex_is(self, regex):
        self._value_equals[self.model_type.name.name_] = re.compile(regex)
        return self
        
    def type_is(self, cls):
        self._value_equals['_type'] = cls._type
        return self

    def type_any_of(self, classes):
        self._value_equals_one_of['_type'] = [cls._type for cls in classes]
        return self
    
    def test_uuid_is(self, test_uuid):
        self._value_equals[self.model_type.test_id.test_uuid.name_] = test_uuid
        return self
    
    def test_uuid_any_of(self, test_uuids):
        self._value_equals_one_of[self.model_type.test_id.test_uuid.name_] = test_uuids
        return self
    
    def testbench_name_is(self, testbench_name):
        self._value_equals[self.model_type.test_id.testbench_name.name_] = testbench_name
        return self
    
    def outcome_is(self, **outcome_attributes):
        for field_name, value in outcome_attributes.items():
            field = getattr(self.model_type.outcome, field_name)
            self._value_equals[field.name_] = value
        return self
    
    def dut_match(self, id=None, version=None):
        if id is not None:
            self._value_equals[self.model_type.test_id.dut.id.name_] = re.compile(id)
        if version is not None:
            self._value_equals[self.model_type.test_id.dut.version.name_] = re.compile(version)
        return self

    def __repr__(self):
        return '{}(eq={}, one-of={}, contains={})'.format(
            self.__class__.__name__,
            repr(self._value_equals),
            repr(self._value_equals_one_of),
            repr(self._list_contains)
        )


import pytest

from datetime import datetime
from spintop import models 
from spintop.models.internal import DutOp, DutIDRecord, NO_VERSION, MeasureFeatureRecord, TestIDRecord, BaseDataClass, DefaultPrimitiveView


from spintop.models.view import ComplexPrimitiveView, DataClassPrimitiveView


def test_simple_tree_test():
    builder = models.SpintopTreeTestRecordBuilder()
    builder.set_top_level_information(
        start_datetime = datetime(year=2020, month=1, day=1),
        dut_id = 'dut_id',
        testbench_name = 'testbench_name',
        outcome=True,
        duration=0.2
    )
    
    with builder.new_phase('test1', True, 0.2) as test1:
        test1.new_measure('test1_measure', True, 4.5)
        
def test_dut_op_strings():
    op = DutOp.create('match', 'new', op_datetime=datetime(year=2019, month=1, day=1))
    dut = DutIDRecord.create('match')
    
    assert op.does_applies_to(dut, on_datetime=datetime(year=2019, month=2, day=1)) # One month later
    assert not op.does_applies_to(dut, on_datetime=datetime(year=2018, month=1, day=1)) # One year previous: does not apply.
    
def test_dut_op_match():
    op = DutOp.create('match', 'new', op_datetime=datetime(year=2019, month=1, day=1))
    match_dut = DutIDRecord.create('match')
    no_match_dut = DutIDRecord.create('no_match')
    
    assert op.does_applies_to(match_dut, on_datetime=datetime(year=2019, month=2, day=1)) # One month later
    assert not op.does_applies_to(no_match_dut, on_datetime=datetime(year=2019, month=2, day=1)) # One month later
    
def test_dut_op_match_version():
    op = DutOp.create({'id': 'match', 'version': 'match_v'}, 'new', op_datetime=datetime(year=2019, month=1, day=1))
    match_dut = DutIDRecord.create('match', version='match_v')
    no_match_dut = DutIDRecord.create('match', version='no_match_v')
    
    assert op.does_applies_to(match_dut, on_datetime=datetime(year=2019, month=2, day=1)) # One month later
    assert not op.does_applies_to(no_match_dut, on_datetime=datetime(year=2019, month=2, day=1)) # One month later
    
    assert op.apply(match_dut).version == NO_VERSION, "The after_dut has no version, the version should be None."

def test_meta_class_field_access():
    assert DutIDRecord.version.name_ == 'version'

def test_nested_field_access():
    assert TestIDRecord.dut.id.name_ == 'dut.id'
    
A_TEST_ID_RECORD = TestIDRecord(
    testbench_name = 'foo',
    dut= DutIDRecord(id='x', version='0'),
    test_uuid= 'unique',
    tags= {'my-tag': True},
    start_datetime= datetime(year=2019, month=2, day=1)
)

def create_a_view(*args, **kwargs):
    primitive_view = ComplexPrimitiveView(BaseDataClass, *args, **kwargs)
    primitive_view.add_view(
        TestIDRecord, 
        {
            'testbench_name': TestIDRecord.testbench_name,
            'dut': TestIDRecord.dut
        }
    )
    primitive_view.add_view(
        DutIDRecord,
        {
            'serial': DutIDRecord.id
        }
    )
    return primitive_view
    
def test_data_primitive_view_deep():
    test_id = A_TEST_ID_RECORD
    
    # primitive_view = ComplexPrimitiveView(cls_missing_view_fn=lambda cls: cls.get_default_view())
    primitive_view = create_a_view()
    
    raw_data = primitive_view.apply(test_id)
    
    assert raw_data == {
        'testbench_name': test_id.testbench_name,
        'dut': {
            'serial': test_id.dut.id
        }
    }
    
def test_data_primitive_view_flatten():
    test_id = A_TEST_ID_RECORD
    
    # primitive_view = ComplexPrimitiveView(cls_missing_view_fn=lambda cls: cls.get_default_view())
    primitive_view = create_a_view()
    
    raw_data = primitive_view.apply(test_id, flatten_dict=True)
    
    assert raw_data == {
        ('testbench_name',): test_id.testbench_name,
        ('dut', 'serial'): test_id.dut.id
    }
    
def test_data_primitive_view_defaults():
    test_id = A_TEST_ID_RECORD
    
    primitive_view = DefaultPrimitiveView()
    raw_data = primitive_view.apply(test_id, flatten_dict=True)
    
    assert raw_data == {
        ('testbench_name',): test_id.testbench_name,
        ('dut', 'id'): test_id.dut.id,
        ('dut', 'version'): test_id.dut.version,
        ('test_uuid',): test_id.test_uuid,
        ('tags', 'my-tag'): test_id.tags['my-tag'],
        ('start_datetime',): test_id.start_datetime,
    }
    
def test_data_primitive_view_deepen():
    test_id = A_TEST_ID_RECORD
    
    # primitive_view = ComplexPrimitiveView(cls_missing_view_fn=lambda cls: cls.get_default_view())
    primitive_view = create_a_view()
    
    raw_data_one_level = primitive_view.apply(test_id, key_prefix=('one',))
    
    assert raw_data_one_level == {
        'one': {
            'testbench_name': test_id.testbench_name,
            'dut': {
                'serial': test_id.dut.id
            }
        }
    }
    
    raw_data_two_levels = primitive_view.apply(test_id, key_prefix=('one', 'two'))
    
    assert raw_data_two_levels == {
        'one': {
            'two': {
                'testbench_name': test_id.testbench_name,
                'dut': {
                    'serial': test_id.dut.id
                }
            }
        }
    }
    
def test_data_primitive_view_flatten_key_prefix():
    test_id = A_TEST_ID_RECORD
    
    # primitive_view = ComplexPrimitiveView(cls_missing_view_fn=lambda cls: cls.get_default_view())
    primitive_view = create_a_view()
    
    raw_data = primitive_view.apply(test_id, flatten_dict=True, key_prefix=('one',))
    
    assert raw_data == {
        ('one', 'testbench_name',): test_id.testbench_name,
        ('one', 'dut', 'serial'): test_id.dut.id
    }
    
def test_data_primitive_key_prefix_callable():
    primitive_view = ComplexPrimitiveView(BaseDataClass)
    primitive_view.add_view(
        DutIDRecord,
        {
            (lambda feat: 'deep',): DutIDRecord.id
        }
    )
    
    dut = DutIDRecord(id='x', version='0')
    raw_data = primitive_view.apply(dut, key_prefix=(lambda feat: 'serial',), flatten_dict=True)
    
    assert raw_data == {
        ('serial', 'deep'): 'x'
    }
    
    